/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "TrayIcon.h"

TrayIcon::TrayIcon()
{
	defaulticon.iconSize = hovericon.iconSize = hiddenicon.iconSize = 16;
}

TrayIcon::~TrayIcon()
{
}

void TrayIcon::subconfigure(string type, ICONSETTINGS &workSettings, ICONSETTINGS &defaultSettings)
{
	workSettings.iconSize = GetxRCInt(type, "IconSize", defaultSettings.iconSize, 8, 64);
	
	workSettings.icon_hueColor = GetxRCColor(type, "IconHueColor", defaultSettings.icon_hueColor);
	workSettings.icon_hueIntensity = GetxRCInt(type, "IconHueIntensity", defaultSettings.icon_hueIntensity, 0, 100);
	workSettings.icon_luminanceIntensity = GetxRCInt(type, "IconLuminanceIntensity", defaultSettings.icon_luminanceIntensity, -100, 100);
	workSettings.icon_saturationIntensity = GetxRCInt(type, "IconSaturationIntensity", defaultSettings.icon_saturationIntensity, 0, 200);

	workSettings.icon_mixColor = GetxRCColor(type, "IconMixColor", defaultSettings.icon_mixColor);
	workSettings.icon_mixIntensity = GetxRCInt(type, "IconMixIntensity", defaultSettings.icon_mixIntensity, 0, 100);
}

void TrayIcon::configure(const string &name)
{
	defaulticon.iconSize = GetxRCInt(name, "IconSize", 16, 8, 64);

	defaulticon.icon_hueColor = GetxRCColor(name, "IconHueColor", RGB(255, 255, 255));
	defaulticon.icon_hueIntensity = GetxRCInt(name, "IconHueIntensity", 0, 0, 100);
	defaulticon.icon_luminanceIntensity = GetxRCInt(name, "IconLuminanceIntensity", 0, -100, 100);
	defaulticon.icon_saturationIntensity = GetxRCInt(name, "IconSaturationIntensity", 100, 0, 200);

	defaulticon.icon_mixColor = GetxRCColor(name, "IconMixColor", RGB(255, 255, 255));
	defaulticon.icon_mixIntensity = GetxRCInt(name, "IconMixIntensity", 0, 0, 100);

	subconfigure((name + "Hover"), hovericon, defaulticon);

	subconfigure((name + "Hidden"), hiddenicon, defaulticon);
}

void TrayIcon::apply(HDC hDC, int x, int y, int width, int height, int buttonType, HICON hIcon, bool useAlphaMap)
{
	if(hIcon == NULL)
		return;

	int iconX = 0;
	int iconY = 0;
	int iconSize = 16;

	COLORREF _hueColor = RGB(255, 255, 255);
	int _hueIntensity = 0;
	int _luminanceIntensity = 0;
	int _saturationIntensity = 100;

	COLORREF _mixColor = RGB(255, 255, 255);
	int _mixIntensity = 0;

	if (buttonType == 0)
	{
		iconSize = defaulticon.iconSize;
		
		_hueColor = defaulticon.icon_hueColor;
		_hueIntensity = defaulticon.icon_hueIntensity;
		_luminanceIntensity = defaulticon.icon_luminanceIntensity;
		_saturationIntensity = defaulticon.icon_saturationIntensity;

		_mixColor = defaulticon.icon_mixColor;
		_mixIntensity = defaulticon.icon_mixIntensity;
	}
	else if (buttonType == 1)
	{
		iconSize = hovericon.iconSize;

		_hueColor = hovericon.icon_hueColor;
		_hueIntensity = hovericon.icon_hueIntensity;
		_luminanceIntensity = hovericon.icon_luminanceIntensity;
		_saturationIntensity = hovericon.icon_saturationIntensity;

		_mixColor = hovericon.icon_mixColor;
		_mixIntensity = hovericon.icon_mixIntensity;
	}
	else if (buttonType == 2)
	{
		iconSize = hiddenicon.iconSize;

		_hueColor = hiddenicon.icon_hueColor;
		_hueIntensity = hiddenicon.icon_hueIntensity;
		_luminanceIntensity = hiddenicon.icon_luminanceIntensity;
		_saturationIntensity = hiddenicon.icon_saturationIntensity;

		_mixColor = hiddenicon.icon_mixColor;
		_mixIntensity = hiddenicon.icon_mixIntensity;
	}

	if ( useAlphaMap || _hueIntensity != 0 || _luminanceIntensity != 0 || _saturationIntensity != 100 || _mixIntensity != 0 )
	{
		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=iconSize;      // Set size you need
		alpha.bmiHeader.biHeight=iconSize;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		HDC hdcBGCOPY = CreateCompatibleDC(hDC);
		HBITMAP	hbmBGCOPY = CreateDIBSection( hdcBGCOPY,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
		HBITMAP oldBGCOPYBMP = (HBITMAP) SelectObject(hdcBGCOPY, hbmBGCOPY);

		//Copy RealBG in Result Background
		BitBlt( hdcBGCOPY, 0, 0, iconSize, iconSize, hDC, x+iconX, y+iconY, SRCCOPY );

		HDC hdcBuffer = CreateCompatibleDC(hDC);
		HBITMAP	hbmBuffer = CreateDIBSection( hdcBuffer,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
		HBITMAP oldBufferBMP = (HBITMAP) SelectObject(hdcBuffer, hbmBuffer);

		//Copy RealBG in Result Background
		BitBlt(hdcBuffer, 0, 0, iconSize, iconSize, hDC, x+iconX, y+iconY, SRCCOPY );

		//Draw Icon
		DrawIconEx(hdcBuffer, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL );

		HDC hdcMask = CreateCompatibleDC(hdcBuffer);
		HBITMAP	hbmMask = CreateDIBSection( hdcMask,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
		HBITMAP oldMaskBMP = (HBITMAP) SelectObject(hdcMask, hbmMask);

		//Draw Mask of the Icon
		DrawIconEx(hdcMask, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_MASK);

		if ( IsOS(OS_XP) )
		{
			BITMAP bm;
			GetObject(hbmBuffer, sizeof(BITMAP), &bm);

			BITMAP bmmask;
			GetObject(hbmMask, sizeof(BITMAP), &bmmask);

			BITMAP bmorg;
			GetObject(hbmBGCOPY, sizeof(BITMAP), &bmorg);

			BYTE *pPixel = (LPBYTE)bm.bmBits;
			BYTE *pPixelMask = (LPBYTE)bmmask.bmBits;
			BYTE *pPixelOrg = (LPBYTE)bmorg.bmBits;

			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;

			BYTE h = 0;
			BYTE s = 0;
			BYTE l = 0;
			
			BYTE hueh = 0;
			BYTE hues = 0;
			BYTE huel = 0;

			bool needHSLConversion = false;
			bool needMixColor = false;

			if ( _luminanceIntensity != 0 || _saturationIntensity != 100 )
				needHSLConversion = true;

			if ( _hueIntensity != 0 )
			{
				r = GetRValue(_hueColor);
				g = GetGValue(_hueColor);
				b = GetBValue(_hueColor);

				if ( r == b && b == g )
				{
					_hueIntensity = 0;
				}
				else
				{
					RGB2HSL(b, g, r, hueh, hues, huel); //Switch r and b due to AlphaMap/bm_b.bmHeight * bm_b.bmWidth
					needHSLConversion = true;
				}
			}

			if ( _mixIntensity != 0 )
			{
				_mixIntensity = (int) min(255, _mixIntensity*2.55);
				needMixColor = true;
			}


			bool skip = false;

			int iSize = bm.bmHeight * bm.bmWidth;

			for (int i = 0; i < iSize; ++i)
			{
				//Filter Mask and insert OriginalBG to preserve AlphaChannel
				if ( pPixelMask[0] == 255 && pPixelMask[1] == 255 && pPixelMask[2] == 255 )
				{
					pPixel[0] = pPixelOrg[0];	
					pPixel[1] = pPixelOrg[1];	
					pPixel[2] = pPixelOrg[2];	
					pPixel[3] = pPixelOrg[3];	

					skip = true;
				}
				//For colorizing without AlphaMap
				if ( !skip && pPixelOrg[0] == 255 && pPixelOrg[1] == 0 && pPixelOrg[2] == 255 && pPixel[3] != 255 && pPixel[3] != 0 )
				{
					pPixel[0] = pPixelOrg[0];	
					pPixel[1] = pPixelOrg[1];	
					pPixel[2] = pPixelOrg[2];	
					pPixel[3] = pPixelOrg[3];	

					skip = true;
				}

				if ( skip )
				{
					pPixel+= 4;
					pPixelMask+= 4;
					pPixelOrg+= 4;

					skip = false;
					continue;
				}

				r = pPixel[0];
				g = pPixel[1];
				b = pPixel[2];

				if ( needHSLConversion )
				{
					//AlphaMap Colorizing
					RGB2HSL(r, g, b, h, s, l);

					if(_hueIntensity != 0)
						h = (BYTE) max(0, min(240, (h*(100-_hueIntensity)+hueh*_hueIntensity)/100));
				
					if(_luminanceIntensity != 0)
						l = (BYTE) max(0, min(240, ((l*100)+(120*_luminanceIntensity))/100));

					if(_saturationIntensity != 100)
					{
						if ( _saturationIntensity > 100 )
							s = (BYTE) max(0, min(240, (240*(_saturationIntensity-100))/100));				
						else
							s = (BYTE) max(0, min(240, (s*_saturationIntensity)/100));
					}

					HSL2RGB(r, g, b, h, s, l);
				}

				if ( needMixColor )
				{
					// HueMix Effect
					if ( pPixel[3] != 0 )
					{
						r = (BYTE)((r * (255 - _mixIntensity) + (pPixel[3]*GetBValue(_mixColor)/255) * _mixIntensity + 255) >> 8);
						g = (BYTE)((g * (255 - _mixIntensity) + (pPixel[3]*GetGValue(_mixColor)/255) * _mixIntensity + 255) >> 8);
						b = (BYTE)((b * (255 - _mixIntensity) + (pPixel[3]*GetRValue(_mixColor)/255) * _mixIntensity + 255) >> 8);
					}
					else
					{
						r = (BYTE)((r * (255 - _mixIntensity) + GetBValue(_mixColor) * _mixIntensity + 255) >> 8);
						g = (BYTE)((g * (255 - _mixIntensity) + GetGValue(_mixColor) * _mixIntensity + 255) >> 8);
						b = (BYTE)((b * (255 - _mixIntensity) + GetRValue(_mixColor) * _mixIntensity + 255) >> 8);
					}
				}

				pPixel[0] = r;
				pPixel[1] = g;
				pPixel[2] = b;
				//Fix for Icons without AlphaChannel (all is 0!), with AlphaChannel 0 is Mask
				if ( pPixel[3] == 0 )
					pPixel[3] = 255;

				pPixel+= 4;
				pPixelMask+= 4;
				pPixelOrg+= 4;
			}

			BitBlt( hDC, x+iconX, y+iconY, iconSize, iconSize, hdcBuffer, 0, 0, SRCCOPY);
		}
		else
		{
			int iy = 0;
			int ix = 0;
			COLORREF cl = 0;
			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;

			BYTE h = 0;
			BYTE s = 0;
			BYTE l = 0;
			
			BYTE hueh = 0;
			BYTE hues = 0;
			BYTE huel = 0;

			bool needHSLConversion = false;
			bool needMixColor = false;

			if ( _luminanceIntensity != 0 || _saturationIntensity != 100 )
				needHSLConversion = true;

			if ( _hueIntensity != 0 )
			{
				r = GetRValue(_hueColor);
				g = GetGValue(_hueColor);
				b = GetBValue(_hueColor);

				if ( r == b && b == g )
				{
					_hueIntensity = 0;
				}
				else
				{
					RGB2HSL(r, g, b, hueh, hues, huel);
					needHSLConversion = true;
				}
			}

			if ( _mixIntensity != 0 )
			{
				_mixIntensity = (int) min(255, _mixIntensity*2.55);
				needMixColor = true;
			}

			for(iy = 0; iy < iconSize; iy++)
			{
				for(ix = 0; ix < iconSize; ix++)
				{
					// Loop through all transparent pixels...
					// If the mask-pixel is white or grey(if that is possible?), then break
					if( !GetPixel(hdcMask, ix, iy) )
					{
						cl = GetPixel(hdcBuffer, ix, iy);
						r = GetRValue(cl);
						g = GetGValue(cl);
						b = GetBValue(cl);

						if ( needHSLConversion )
						{
							RGB2HSL(r, g, b, h, s, l);

							if(_hueIntensity != 0)
								h = (BYTE) max(0, min(240, (h*(100-_hueIntensity)+hueh*_hueIntensity)/100));
				
							if(_luminanceIntensity != 0)
								l = (BYTE) max(0, min(240, ((l*100)+(120*_luminanceIntensity))/100));

							if(_saturationIntensity != 100)
							{
								if ( _saturationIntensity > 100 )
									s = (BYTE) max(0, min(240, (240*(_saturationIntensity-100))/100));				
								else
									s = (BYTE) max(0, min(240, (s*_saturationIntensity)/100));
							}
						
							HSL2RGB(r, g, b, h, s, l);
						}

						if ( needMixColor )
						{
							// HueMix Effect
							r = (BYTE)((r * (255 - _mixIntensity) + GetRValue(_mixColor) * _mixIntensity + 255) >> 8);
							g = (BYTE)((g * (255 - _mixIntensity) + GetGValue(_mixColor) * _mixIntensity + 255) >> 8);
							b = (BYTE)((b * (255 - _mixIntensity) + GetBValue(_mixColor) * _mixIntensity + 255) >> 8);
						}

						SetPixelV(hdcBGCOPY, ix, iy, RGB(r,g,b));
					}
				}
			}

			BitBlt( hDC, x+iconX, y+iconY, iconSize, iconSize, hdcBGCOPY, 0, 0, SRCCOPY);
		}
				
		hbmMask = (HBITMAP) SelectObject(hdcMask, oldMaskBMP);
		DeleteObject(hbmMask);
		DeleteDC(hdcMask);

		hbmBuffer = (HBITMAP) SelectObject(hdcBuffer, oldBufferBMP);
		DeleteObject(hbmBuffer);
		DeleteDC(hdcBuffer);

		hbmBGCOPY = (HBITMAP) SelectObject(hdcBGCOPY, oldBGCOPYBMP);
		DeleteObject(hbmBGCOPY);
		DeleteDC(hdcBGCOPY);
	}
	else
	{
		DrawIconEx(hDC, x+iconX, y+iconY, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);
	}
}
