/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Utilities.h"

#define INVALID_ICON(x) ((NULL == (x)) || (1 == (int)(x)) || (-1 == (int)(x)))

HDC hdcDesktop = 0;
HBITMAP hbmDesktop = 0;

void PrefixedLSSetVariable(string prefix, string var, string value)
{
	if ( !var.empty() )
		prefix.append(var);

	LSSetVariable(prefix.c_str(), value.c_str());
}

string GetTaskClass( HWND hTask )
{
	char windowClass[MAX_LINE_LENGTH] = {0};

	if (hTask)
		GetClassName( hTask, windowClass, MAX_LINE_LENGTH );

	return (string) windowClass;
}

string GetTaskWindowText( HWND hTask )
{
	char windowText[MAX_LINE_LENGTH] = {0};

	if (hTask)
		GetWindowText( hTask, windowText, MAX_LINE_LENGTH );

	return (string) windowText;
}

bool IsSubString(string big, string small)
{
	if( match( small.c_str(), big.c_str() ) )
		return true;
	else
		return false;
}

bool IsOnTrayList(HWND hWnd, vector<CLASS_CONFIG> theList, bool emptyValue)
{	
	//Matching Lists
	if ( theList.empty() )
		return emptyValue;

	string tmpClass = GetTaskClass( hWnd );
	string tmpWindow = GetTaskWindowText( hWnd );

	if ( tmpClass.empty() )
		return emptyValue;
	
	for (vector<CLASS_CONFIG>::iterator it = theList.begin(); it != theList.end(); ++it)
	{
		if ( IsSubString(tmpClass, (*it).classname) )
		{
			if ( !(*it).windowname.empty() )
			{
				if ( !IsSubString(tmpWindow, (*it).windowname) )
					continue;
			}

			return !emptyValue;
		}
	}

	return emptyValue;
}

int PosOnTrayList(HWND hWnd, vector<CLASS_CONFIG> theList, int emptyValue)
{	
	//Matching Lists
	if ( theList.empty() )
		return emptyValue;

	string tmpClass = GetTaskClass( hWnd );
	string tmpWindow = GetTaskWindowText( hWnd );

	if ( tmpClass.empty() )
		return emptyValue;
	
	for (vector<CLASS_CONFIG>::iterator it = theList.begin(); it != theList.end(); ++it)
	{
		if ( IsSubString(tmpClass, (*it).classname) )
		{
			if ( !(*it).windowname.empty() )
			{
				if ( !IsSubString(tmpWindow, (*it).windowname) )
					continue;
			}

			return (*it).position;
		}
	}

	return emptyValue;
}

bool IsOS(const int &os)
{	
	static int osversion = 0;

	static bool bInitialized = false;

	if (!bInitialized)
	{
		OSVERSIONINFO verInfo;
		verInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		GetVersionEx(&verInfo);
		
		switch(verInfo.dwPlatformId)
		{
			case VER_PLATFORM_WIN32_WINDOWS:

				osversion |= OS_9X;
			
				if(verInfo.dwMinorVersion >= 90)
				{
					osversion |= OS_ME;
				}
				else if(verInfo.dwMinorVersion >= 10)
				{
					osversion |= OS_98;
				}
				else
				{
					osversion |= OS_95;
				}
				break;
			
			case VER_PLATFORM_WIN32_NT:
			
				osversion |= OS_NT;
			
				if(verInfo.dwMajorVersion >= 5)
				{
					osversion |= OS_2KXP;

					if (verInfo.dwMinorVersion >= 1)
					{
						osversion |= OS_XP;
					}
					else
					{
						osversion |= OS_2K;
					}
				}
				else
				{
					osversion |= OS_NT4;
				}
				break;
			default:;
		}
		bInitialized = true;
	}

	return (osversion & os) ? true : false;
}

string QueryPixelDepth()
{
	CString Response = "";
	DEVMODE DevMode;
	DevMode.dmSize=sizeof(DEVMODE);
	
	if (EnumDisplaySettings(NULL,ENUM_CURRENT_SETTINGS,&DevMode))
	{
		Response.Format("%ld", DevMode.dmBitsPerPel);
	}
	else
	{
		Response = "8";
	}
	
	return (string)Response;
}

// retrieve handle of the desktop window
HWND GetLitestepDesktop()
{
	HWND hWnd = FindWindow("DesktopBackgroundClass", NULL);

	if(hWnd != 0)
		return hWnd;

	return GetDesktopWindow();
}

HWND FindWindowEx2(string pszClass, string pszTitle)
{
	HWND hWnd = NULL;
	FindWindowEx2Data fwed;
	fwed.phWnd = &hWnd;
	fwed.pszClass = pszClass.c_str();
	fwed.pszTitle = pszTitle.c_str();
	EnumWindows(FindWindowEx2EnumWindowsProc, (LPARAM)&fwed);
	return *(fwed.phWnd);
}

BOOL CALLBACK FindWindowEx2EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
	FindWindowEx2Data* pfwed = (FindWindowEx2Data*)lParam;
	char szClass[1024];
	char szTitle[1024];
	if(pfwed->pszClass)
	{
		GetClassName(hWnd, szClass, 1024);
	}
	if(pfwed->pszTitle)
	{
		GetWindowText(hWnd, szTitle, 1024);
	}
	if((pfwed->pszClass != NULL ? !_stricmp(szClass, pfwed->pszClass) : TRUE) && (pfwed->pszTitle != NULL ? !_stricmp(szTitle, pfwed->pszTitle) : TRUE))
	{
		*(pfwed->phWnd) = hWnd;
		return FALSE;
	}
	EnumChildWindows(hWnd, FindWindowEx2EnumWindowsProc, lParam);

	return (*pfwed->phWnd == NULL);
}

// modify window style
void ModifyStyle(HWND hWnd, DWORD removeStyle, DWORD addStyle)
{
	SetWindowLong(hWnd, GWL_STYLE, (GetWindowLong(hWnd, GWL_STYLE) & ~removeStyle) | addStyle);
}

void ModifyStyleEx(HWND hWnd, DWORD removeStyle, DWORD addStyle)
{
	SetWindowLong(hWnd, GWL_EXSTYLE, (GetWindowLong(hWnd, GWL_EXSTYLE) & ~removeStyle) | addStyle);
}

// trim leading and trailing whitespace
string trim(const string &toTrim)
{
	int start = 0;
	int end = toTrim.length() - 1;
	
	while(start <= end && isspace(toTrim[start]))
		start++;
	
	while(start <= end && isspace(toTrim[end]))
		end--;
	
	return toTrim.substr(start, end - start + 1);
}

int SplitString(const string &process, string &one, string &two, string &three, string &four, string &five, string &six, string &seven)
{
	int count = 0;
	one = "";
	two = "";
	three = "";
	four = "";
	five = "";
	six = "";
	seven = "";
	char cone[256] = {0};
	char ctwo[256] = {0};
	char cthree[256] = {0};
	char cfour[256] = {0};
	char cfive[256] = {0};
	char csix[256] = {0};
	char cseven[256] = {0};
	char *cbuffers[] = {cone, ctwo, cthree, cfour, cfive, csix, cseven};
		
	int numTokens = LCTokenize( process.c_str(), cbuffers, 7, NULL);
	if( numTokens >= 1 )
	{
		one = cone;
		count++;
		
		if ( numTokens >= 2 )
		{
			two = ctwo;
			count++;
		}
		
		if ( numTokens >= 3 )
		{
			three = cthree;
			count++;
		}
		
		if ( numTokens >= 4 )
		{
			four = cfour;
			count++;
		}

		if ( numTokens >= 5 )
		{
			five = cfive;
			count++;
		}

		if ( numTokens >= 6 )
		{
			six = csix;
			count++;
		}

		if ( numTokens >= 7 )
		{
			seven = cseven;
			count++;
		}
	}

	return count;
}

StringList GetxRCLineList(const string &uniqueKey)
{
	StringList resultList;
	LPVOID f = LCOpen(NULL);

	if(f != NULL)
	{
		char lineBuffer[MAX_LINE_LENGTH];

		while(LCReadNextConfig(f, uniqueKey.c_str(), lineBuffer, MAX_LINE_LENGTH))
		{
			char temp[64], configline[MAX_LINE_LENGTH];
			char *buffers[] = {temp};
			
			if(LCTokenize(lineBuffer, buffers, 1, configline) >= 1)
			{
				resultList.insert(resultList.end(), configline);
			}
		}
		
		LCClose(f);
	}

	return resultList;
}

void GetxRCList(char *text, vector<CLASS_CONFIG> &theList, bool setPosition)
{
	LPVOID f = LCOpen(NULL);

	if(f != NULL)
	{
		char lineBuffer[MAX_LINE_LENGTH] = {0};

		while(LCReadNextConfig(f, text, lineBuffer, MAX_LINE_LENGTH))
		{
			char temp[64] = {0};
			char first[MAX_LINE_LENGTH] = {0};
			char second[MAX_LINE_LENGTH] = {0};
			char third[MAX_LINE_LENGTH] = {0};
			char *buffers[] = {temp, first, second, third};

			CLASS_CONFIG hideCFG;
			
			int numTokens = LCTokenize(lineBuffer, buffers, 4, NULL);

			if ( !setPosition && numTokens == 2 )
			{
				hideCFG.classname = first;
				hideCFG.windowname = "";
				hideCFG.position = -1;

				if ( !hideCFG.classname.empty() )
					theList.insert(theList.end(), hideCFG);
			}
			else if ( !setPosition && numTokens == 3 )
			{
				hideCFG.classname = first;
				hideCFG.windowname = second;
				hideCFG.position = -1;

				if ( !hideCFG.classname.empty() )
					theList.insert(theList.end(), hideCFG);
			}
			else if ( setPosition && numTokens == 3 )
			{
				hideCFG.position = atoi(first);
				hideCFG.classname = second;
				hideCFG.windowname = "";
				
				if ( !hideCFG.classname.empty() )
					theList.insert(theList.end(), hideCFG);
			}
			else if ( setPosition && numTokens == 4 )
			{	
				hideCFG.position = atoi(first);
				hideCFG.classname = second;
				hideCFG.windowname = third;
			
				if ( !hideCFG.classname.empty() )
					theList.insert(theList.end(), hideCFG);
			}
		}
		
		LCClose(f);
	}
}

// retrieve a boolean value from the step.rc
bool GetxRCBool(const string &object, const string &subKey, bool defaultVal)
{
	string key = object + subKey;
	return GetRCBoolDef(key.c_str(), defaultVal) ? true : false;
}

// retrieve a string value from step.rc
string GetxRCString(const string &object, const string &subKey, const string &defaultVal)
{
	string key = object + subKey;
	char buffer[MAX_LINE_LENGTH];
	GetRCString(key.c_str(), buffer, defaultVal.c_str(), MAX_LINE_LENGTH);
	return string(buffer);
}

// retrieve a line from step.rc
string GetxRCLine(const string &object, const string &subKey, const string &defaultVal)
{
	string key = object + subKey;
	char buffer[MAX_LINE_LENGTH];
	GetRCLine(key.c_str(), buffer, MAX_LINE_LENGTH, defaultVal.c_str());
	return (string) buffer;
}

// retrieve a named value from step.rc
int GetxRCNamedValue(const string &prefix, const string &baseName, const NameValuePair *nameValuePairs, int defaultVal)
{
	string name = GetxRCString(prefix, baseName, "");
	const NameValuePair *current = nameValuePairs;

	if(name.length() > 0)
	{
		while(current->name)
		{
			if(_stricmp(current->name, name.c_str()) == 0)
				return current->value;

			current++;
		}
	}

	return defaultVal;
}

// retrieve a color value from step.rc
int GetxRCColor(const string &object, const string &subKey, int defaultVal)
{
	string key = object + subKey;

	return GetRCColor(key.c_str(), defaultVal);
}

// retrieve an integer value from step.rc, make sure it is between minVal and maxVal inclusive
int GetxRCInt(const string &object, const string &subKey, int defaultVal, int minVal, int maxVal)
{
	string key = object + subKey;
	int answer = GetRCInt(key.c_str(), defaultVal);
	return max(min(answer, maxVal), minVal);
}

// parse a coordinate specification
int xParseCoord(const string &aString, int defaultVal, int maxVal)
{
	string strVal = trim(aString);

	if(strVal.length() == 0)
		return defaultVal;

	bool negative = false;
	bool realnegative = false;
	bool center = false;
	UINT i = 0;
	int value = 0;

	if(strVal[i] == '-')
	{
		negative = true;
		i++;
	}
	else if(strVal[i] == '~')
	{
		realnegative = true;
		i++;
	}
	else if(strVal[i] == '+')
	{
		i++;
	}

	while(i < strVal.length())
	{
		if(strVal[i] >= '0' && strVal[i] <= '9')
		{
			value = (value * 10) + (strVal[i] - '0');
		}
		else
		{
			if(strVal[i] == 'c' || strVal[i] == 'C')
				center = true;

			break;
		}

		i++;
	}

	if(center)
	{
		if (realnegative)
			negative = true;

		if(negative)
			value = (maxVal / 2) - value;
		else
			value = (maxVal / 2) + value;
	}
	else if(realnegative)
	{
		value = -value;
	}
	else if(negative)
	{
		value = maxVal - value;
	}

	return value;
}

// retrieve a coordinate from step.rc, accounting for negative values and centering
int GetxRCCoord(const string &prefix, const string &baseName, int defaultVal, int maxVal)
{
	string strVal = GetxRCString(prefix, baseName, "");
	return xParseCoord(strVal, defaultVal, maxVal);
}

// parse a dimension
int xParseDimen(const string &aString, int defaultVal, int maxVal)
{
	string strVal = trim(aString);

	if(strVal.length() == 0)
		return defaultVal;

	bool negative = false;
	bool percentage = false;
	bool extendedpercentage = false;
	UINT i = 0;
	int value = 0;
	int extendedvalue = 0;

	if(strVal[i] == '-')
	{
		negative = true;
		i++;
	}

	while(i < strVal.length())
	{
		if(strVal[i] >= '0' && strVal[i] <= '9')
		{
			value = (value * 10) + (strVal[i] - '0');
		}
		else
		{
			if(strVal[i] == '%')
				percentage = true;

			if (strVal[i+1] == '+' )
			{
				extendedpercentage = true;
				i++;

				while (i < strVal.length())
				{
					if (strVal[i] >= '0' && strVal[i] <= '9')
					{
						extendedvalue = (extendedvalue * 10) + (strVal[i] - '0');
					}
					i++;
				}
			}
			else if (strVal[i+1] == '-' )
			{
				extendedpercentage = true;
				i++;

				while (i < strVal.length())
				{
					if (strVal[i] >= '0' && strVal[i] <= '9')
					{
						extendedvalue = (extendedvalue * 10) + (strVal[i] - '0');
					}
					i++;
				}
				extendedvalue = extendedvalue*(-1);
			}

			break;
		}

		i++;
	}

	if(percentage)
	{
		value = (value * maxVal) / 100;
			
		if (extendedpercentage)
			value = value + extendedvalue;
	}

	if(negative)
		value = maxVal - value;

	return value;
}

// retrieve a dimension (absolute value or percentage) from step.rc
int GetxRCDimen(const string &prefix, const string &baseName, int defaultVal, int maxVal)
{
	string strVal = GetxRCString(prefix, baseName, "");
	return xParseDimen(strVal, defaultVal, maxVal);
}

// enhanced version of PaintDesktop
void PaintDesktopEx(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, int xSrc, int ySrc, BOOL updateCache)
{
	if(hdcDesktop == 0 || updateCache)
	{
		if(hdcDesktop == 0 && hdcDest == 0)
			return;

		if(hbmDesktop != 0)
		{
			hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
			DeleteObject(hbmDesktop);
			hbmDesktop = 0;
		}

		HDC hdcScreen = GetDC(0);

		if(hdcDesktop == 0)
		{
			hdcDesktop = CreateCompatibleDC(hdcScreen);
		}
		
		int cxDesktop = GetSystemMetrics(SM_CXSCREEN);
		int cyDesktop = GetSystemMetrics(SM_CYSCREEN);

		hbmDesktop = CreateCompatibleBitmap(hdcScreen, cxDesktop, cyDesktop);
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);

		RECT r;
		SetRect(&r, 0, 0, cxDesktop, cyDesktop);
		FillRect(hdcDesktop, &r, GetSysColorBrush(COLOR_DESKTOP));

		char wallpaper[MAX_LINE_LENGTH] = { 0 };

		RegQueryStringValue(HKEY_CURRENT_USER,
			"Control Panel\\Desktop", "Wallpaper", wallpaper,
			MAX_LINE_LENGTH, "");

		if(wallpaper[0])
		{
			int wallpaperStyle;
			int tileWallpaper;

			char buffer[32];

			RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "WallpaperStyle", buffer, 32, "0");
			wallpaperStyle = atoi(buffer);

			RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "TileWallpaper", buffer, 32, "0");
			tileWallpaper = atoi(buffer);

			HBITMAP hbmWallpaper = (HBITMAP) LoadImage(0, wallpaper,
				IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);

			if(hbmWallpaper)
			{
				BITMAP bm;
				GetObject(hbmWallpaper, sizeof(BITMAP), &bm);

				int cxWallpaper = bm.bmWidth;
				int cyWallpaper = bm.bmHeight;

				HDC hdcWallpaper = CreateCompatibleDC(hdcScreen);
				hbmWallpaper = (HBITMAP) SelectObject(hdcWallpaper, hbmWallpaper);

				if(tileWallpaper)
				{
					TileBlt(hdcDesktop,
						0, 0,
						cxDesktop, cyDesktop,
						hdcWallpaper,
						0, 0,
						cxWallpaper, cyWallpaper,
						SRCCOPY);
				}
				else
				{
					if(wallpaperStyle == 2) // stretch
					{
						SetStretchBltMode(hdcDesktop, STRETCH_DELETESCANS);

						StretchBlt(hdcDesktop,
							0, 0,
							cxDesktop, cyDesktop,
							hdcWallpaper,
							0, 0,
							cxWallpaper, cyWallpaper,
							SRCCOPY);
					}
					else
					{
						BitBlt(hdcDesktop,
						(cxDesktop - cxWallpaper) / 2,
						(cyDesktop - cyWallpaper) / 2,
						cxWallpaper, cyWallpaper,
						hdcWallpaper,
						0, 0,
						SRCCOPY);
					}
				}

				hbmWallpaper = (HBITMAP) SelectObject(hdcWallpaper, hbmWallpaper);
				DeleteDC(hdcWallpaper);

				DeleteObject(hbmWallpaper);
			}
		}

		ReleaseDC(0, hdcScreen);
	}

	if(hdcDest)
	{
		BitBlt(hdcDest, xDest, yDest, cxDest, cyDest, hdcDesktop, xSrc, ySrc, SRCCOPY);
	}
}

// tile an image horizontally and vertically
void TileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp)
{
	for(int y = 0; y < cyDest; y += cySrc)
	{
		for(int x = 0; x < cxDest; x += cxSrc)
		{
			BitBlt(hdcDest,
				xDest + x,
				yDest + y,
				min(cxDest - x, cxSrc),
				min(cyDest - y, cySrc),
				hdcSrc,
				xSrc,
				ySrc,
				rasterOp);
		}
	}
}

// tile an image horizontally
void HorizontalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp)
{
	for(int x = 0; x < cxDest; x += cxSrc)
	{
		StretchBlt(hdcDest,
			xDest + x,
			yDest,
			min(cxDest - x, cxSrc),
			cyDest,
			hdcSrc,
			xSrc,
			ySrc,
			min(cxDest - x, cxSrc),
			cySrc,
			rasterOp);
	}
}

// tile an image vertically
void VerticalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp)
{
	for(int y = 0; y < cyDest; y += cySrc)
	{
		StretchBlt(hdcDest,
			xDest,
			yDest + y,
			cxDest,
			min(cyDest - y, cySrc),
			hdcSrc,
			xSrc,
			ySrc,
			cxSrc,
			min(cyDest - y, cySrc),
			rasterOp);
	}
}

// multi-blt!
void MultiBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, int cxLeft, int cyTop, int cxRight, int cyBottom, int mode, DWORD rasterOp)
{
	if(cyTop > 0)
	{
		if(cxLeft > 0)
		{
			BitBlt(hdcDest,
				xDest,
				yDest,
				cxLeft,
				cyTop,
				hdcSrc,
				xSrc,
				ySrc,
				rasterOp);
		}

		if(mode == MULTIBLT_STRETCH || mode == MULTIBLT_TILEVERTICAL)
		{
			StretchBlt(hdcDest,
				xDest + cxLeft,
				yDest,
				cxDest - cxLeft - cxRight,
				cyTop,
				hdcSrc,
				xSrc + cxLeft,
				ySrc,
				cxSrc - cxLeft - cxRight,
				cyTop,
				rasterOp);
		}
		else
		{
			HorizontalTileBlt(hdcDest,
				xDest + cxLeft,
				yDest,
				cxDest - cxLeft - cxRight,
				cyTop,
				hdcSrc,
				xSrc + cxLeft,
				ySrc,
				cxSrc - cxLeft - cxRight,
				cyTop,
				rasterOp);
		}

		if(cxRight > 0)
		{
			BitBlt(hdcDest,
				cxDest - cxRight,
				yDest,
				cxRight,
				cyTop,
				hdcSrc,
				cxSrc - cxRight,
				ySrc,
				rasterOp);
		}
	}

	if(cxLeft > 0)
	{
		if(mode == MULTIBLT_STRETCH || mode == MULTIBLT_TILEHORIZONTAL)
		{
			StretchBlt(hdcDest,
				xDest,
				yDest + cyTop,
				cxLeft,
				cyDest - cyTop - cyBottom,
				hdcSrc,
				xSrc,
				ySrc + cyTop,
				cxLeft,
				cySrc - cyTop - cyBottom,
				rasterOp);
		}
		else
		{
			VerticalTileBlt(hdcDest,
				xDest,
				yDest + cyTop,
				cxLeft,
				cyDest - cyTop - cyBottom,
				hdcSrc,
				xSrc,
				ySrc + cyTop,
				cxLeft,
				cySrc - cyTop - cyBottom,
				rasterOp);
		}
	}

	if(mode == MULTIBLT_STRETCH)
	{
		StretchBlt(hdcDest,
			xDest + cxLeft,
			yDest + cyTop,
			cxDest - cxLeft - cxRight,
			cyDest - cyTop - cyBottom,
			hdcSrc,
			xSrc + cxLeft,
			ySrc + cyTop,
			cxSrc - cxLeft - cxRight,
			cySrc - cyTop - cyBottom,
			rasterOp);
	}
	else if(mode == MULTIBLT_TILE)
	{
		TileBlt(hdcDest,
			xDest + cxLeft,
			yDest + cyTop,
			cxDest - cxLeft - cxRight,
			cyDest - cyTop - cyBottom,
			hdcSrc,
			xSrc + cxLeft,
			ySrc + cyTop,
			cxSrc - cxLeft - cxRight,
			cySrc - cyTop - cyBottom,
			rasterOp);
	}
	else if(mode == MULTIBLT_TILEHORIZONTAL)
	{
		HorizontalTileBlt(hdcDest,
			xDest + cxLeft,
			yDest + cyTop,
			cxDest - cxLeft - cxRight,
			cyDest - cyTop - cyBottom,
			hdcSrc,
			xSrc + cxLeft,
			ySrc + cyTop,
			cxSrc - cxLeft - cxRight,
			cySrc - cyTop - cyBottom,
			rasterOp);
	}
	else if(mode == MULTIBLT_TILEVERTICAL)
	{
		VerticalTileBlt(hdcDest,
			xDest + cxLeft,
			yDest + cyTop,
			cxDest - cxLeft - cxRight,
			cyDest - cyTop - cyBottom,
			hdcSrc,
			xSrc + cxLeft,
			ySrc + cyTop,
			cxSrc - cxLeft - cxRight,
			cySrc - cyTop - cyBottom,
			rasterOp);
	}

	if(cxRight > 0)
	{
		if(mode == MULTIBLT_STRETCH || mode == MULTIBLT_TILEHORIZONTAL)
		{
			StretchBlt(hdcDest,
				cxDest - cxRight,
				yDest + cyTop,
				cxRight,
				cyDest - cyTop - cyBottom,
				hdcSrc,
				cxSrc - cxRight,
				ySrc + cyTop,
				cxRight,
				cySrc - cyTop - cyBottom,
				rasterOp);
		}
		else
		{
			VerticalTileBlt(hdcDest,
				cxDest - cxRight,
				yDest + cyTop,
				cxRight,
				cyDest - cyTop - cyBottom,
				hdcSrc,
				cxSrc - cxRight,
				ySrc + cyTop,
				cxRight,
				cySrc - cyTop - cyBottom,
				rasterOp);
		}
	}

	if(cyBottom > 0)
	{
		if(cxLeft > 0)
		{
			BitBlt(hdcDest,
				xDest,
				cyDest - cyBottom,
				cxLeft,
				cyBottom,
				hdcSrc,
				xSrc,
				cySrc - cyBottom,
				rasterOp);
		}

		if(mode == MULTIBLT_STRETCH || mode == MULTIBLT_TILEVERTICAL)
		{
			StretchBlt(hdcDest,
				xDest + cxLeft,
				cyDest - cyBottom,
				cxDest - cxLeft - cxRight,
				cyBottom,
				hdcSrc,
				xSrc + cxLeft,
				cySrc - cyBottom,
				cxSrc - cxLeft - cxRight,
				cyBottom,
				rasterOp);
		}
		else
		{
			HorizontalTileBlt(hdcDest,
				xDest + cxLeft,
				cyDest - cyBottom,
				cxDest - cxLeft - cxRight,
				cyBottom,
				hdcSrc,
				xSrc + cxLeft,
				cySrc - cyBottom,
				cxSrc - cxLeft - cxRight,
				cyBottom,
				rasterOp);
		}

		if(cxRight > 0)
		{
			BitBlt(hdcDest,
				cxDest - cxRight,
				cyDest - cyBottom,
				cxRight,
				cyBottom,
				hdcSrc,
				cxSrc - cxRight,
				cySrc - cyBottom,
				rasterOp);
		}
	}
}

// retrieve a string value from the registry
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue)
{
	DWORD dwDataType;
	DWORD dwDataSize = (DWORD) nBufferSize;
	HKEY hSubKey = NULL;
	LONG lErrorCode;
	
	if(pszSubKey)
	{
		lErrorCode = RegOpenKeyEx(hKey, pszSubKey, 0, KEY_QUERY_VALUE, &hSubKey);
		
		if(lErrorCode != ERROR_SUCCESS)
		{
			return pszBuffer;
		}
		
		hKey = hSubKey;
	}
	
	lErrorCode = RegQueryValueEx(hKey, pszValueName, NULL, &dwDataType,
		(LPBYTE) pszBuffer, &dwDataSize);
	
	if(hSubKey)
		RegCloseKey(hSubKey);
	
	return pszBuffer;
}

string IntToString(int input)
{
	char buffer[32];
	itoa( input, buffer, 10 );
	return (string) buffer;
}

void RGB2HSL(BYTE& R, BYTE& G, BYTE& B, BYTE& H, BYTE& S, BYTE& L)
{
	R = min(max(0,R),255);
	G = min(max(0,G),255);
	B = min(max(0,B),255);
	BYTE minval = min(min(R,G),B);
	BYTE maxval = max(max(R,G),B);
	DWORD msum = maxval+minval;
	L = (BYTE)(0.5+(msum*240/510.));

	if (maxval == minval)
	{
		S = 0;
		H = 160;
	}
	else
	{
		float delta = (float)maxval-minval;

		S = (BYTE)(0.5+(240*(delta/((msum<=255) ? msum : (510-msum)))));

		float h;
        
		if (R == maxval)
			h = 40*((G-B)/delta);
		else if (G == maxval)
			h = 40*(2+(B-R)/delta);
		else if (B == maxval)
			h = 40*(4+(R-G)/delta);

		if (h > 240)
			h -= 240;

		H = (BYTE)(0.5+h);
	}
}

void HSL2RGB(BYTE& R, BYTE& G, BYTE& B, BYTE& _H, BYTE& _S, BYTE& _L)
{
	_H = min(max(0,_H),239);
	_S = min(max(0,_S),240);

	_L = min(max(0,_L),240);
	
	float H = _H*360/240.0f;
	float S = _S/240.0f;
	float L = _L/240.0f;

	if (S)
	{
		float rm2;

		if (_L<=120)
			rm2 = L+L*S;  
		else
			rm2 = L+S-L*S;
		float rm1 = 2*L-rm2;   
		
		R = ToRGB(rm1, rm2, H+120);   
		G = ToRGB(rm1, rm2, H);
		B = ToRGB(rm1, rm2, H-120);
	}
	else 
	{
		R = G = B = (BYTE)(0.5+L*255);
	}
}

BYTE ToRGB(float rm1, float rm2, float rh)
{
    if (rh > 360)
		rh -= 360;
	else if (rh < 0) 
		rh += 360;

    if (rh < 60)
		rm1 = rm1+(rm2-rm1)*rh/60;   
	else if (rh < 180)
		rm1 = rm2;
	else if (rh < 240)
		rm1 = rm1+(rm2-rm1)*(240-rh)/60;      

	return (BYTE)(0.5+rm1*255);
}